use async_io::{Async, ReadableOwned};
use futures::prelude::*;
use futures_lite::ready;
use nix::{fcntl::{fcntl, OFlag, F_GETFL, F_SETFL}, unistd::read};
use std::{io, os::{fd::{AsFd, BorrowedFd}, unix::io::AsRawFd}, pin::Pin, sync::Arc, task::{Context, Poll}};

struct XcbAsFdWrapper(xcb::Connection);

impl AsFd for XcbAsFdWrapper {
    fn as_fd(&self) -> std::os::unix::prelude::BorrowedFd<'_> {
        let fd = self.0.as_raw_fd();
        if fd < 0 {
            panic!("XCB file descriptor is invalid");
        }
        unsafe { BorrowedFd::borrow_raw(fd) }
    }
}

pub struct AsyncConnection {
    conn: Arc<Async<XcbAsFdWrapper>>,
    readable: Option<ReadableOwned<XcbAsFdWrapper>>,
}

impl AsyncConnection {
    pub fn new(conn: xcb::Connection) -> io::Result<Self> {
        let flags = fcntl(conn.as_raw_fd(), F_GETFL)?;
        fcntl(conn.as_raw_fd(), F_SETFL(OFlag::from_bits_truncate(flags) | OFlag::O_NONBLOCK))?;
        Ok(Self {
            conn: Arc::new(Async::new(XcbAsFdWrapper(conn))?),
            readable: None,
        })
    }
}

impl AsyncRead for AsyncConnection {
    fn poll_read(mut self: Pin<&mut Self>, cx: &mut Context<'_>, buf: &mut [u8]) -> Poll<std::io::Result<usize>> {
        loop {
            match read(self.conn.as_fd().as_raw_fd(), buf) {
                Err(nix::Error::EAGAIN) => (),
                Err(err) => {
                    self.readable = None;
                    return Poll::Ready(Err(err.into()));
                },
                Ok(count) => {
                    self.readable = None;
                    return Poll::Ready(Ok(count));
                }
            }

            if self.readable.is_none() {
                self.readable = Some(Arc::clone(&self.conn).readable_owned());
            }

            if let Some(f) = &mut self.readable {
                let res = ready!(Pin::new(f).poll(cx));
                self.readable = None;
                res?;
            }
        }
    }
}
